/*
 * grid.c:
 * Conversions between geographical coordinates and Ordnance Survey National
 * Grid or Irish Grid.
 *
 * This file is part of gpstrans - a program to communicate with garmin gps
 * Parts are taken from John F. Waers (jfwaers@csn.net) program MacGPS.
 *
 *    Copyright (c) 1995 by Carsten Tschach (tschach@zedat.fu-berlin.de)
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for non-commercial purpose, is hereby granted without fee,
 * providing that the copyright notice appear in all copies and that both
 * the copyright notice and this permission notice appear in supporting
 * documentation. I make no representations about the suitability of this
 * software for any purpose. It is provides "as is" without express or
 * implied warranty.
 *
 * Irish Grid support added by Chris Lightfoot <chris@ex-parrot.com>.
 *
 * $Id: grid.cc,v 1.3 2001/12/06 17:47:10 root Exp $
 *
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "coords.h"

/* Conversion to/from OSGB grid. */

/* osgb_ constants:
 * These define the OSGB `National Grid'.
 *
 * The osgb_k0 constant was empirically determined to give the best fit to the
 * data from the GPS 45 and the PCX5 program. It should be 0.9996, but this
 * number resulted in slight errors in the data.
 */
//static const char *osgb_zoneID0 = "STUQRNOPLMHJKFGCDEABXYZVW";
//static const char *osgb_zoneID1 = "VWXYZQRSTULMNOPFGHJKABCDE";
static const double osgb_lat0 = 49.0; /* reference transverse mercator latitude */
static const double osgb_lon0 = -2.0; /* and longitude */
static const double osgb_k0 = 0.99960133; /* XXX shd be 0.996012717 */

static const double osgb_f_e = 400000.; /* Map coordinates of true origin. */
static const double osgb_f_n = -100000.;

/* geog_to_osgb:
/* Convert geographical coordinates in the given datum into an Ordnance Survey
 * zone and eastings/northings.
 */
int
geog_to_osgb (double lat, double lon, double *x, double *y, int datum)
{

  geog_to_tm (lat, lon, osgb_lat0, osgb_lon0, osgb_k0, x, y, datum);

  *x = *x + osgb_f_e;
  *y = *y + osgb_f_n;

  /* check for invalid range */
  if (*x < 0.0 || *x > 1.0e6 || *y < 0.0 || *y > 2.5e6)
    {
      *x = 0.0;
      *y = 0.0;

      return 0;
    }

  return 1;
}


/* osgb_to_geog:
 * Convery Ordnance Survey coordinates into geographical coordinates in the
 * given datum.
 */
int
osgb_to_geog (double x, double y, double *lat, double *lon, int datum)
{

  x -= osgb_f_e;
  y -= osgb_f_n;

  tm_to_geog (x, y, osgb_lat0, osgb_lon0, osgb_k0, lat, lon, datum);

  return 1;
}

/* Conversion to/from Irish Grid. */

/* irish_ constants:
 * These define the Irish Grid.
 */
//static const char *irish_zoneID = "VWXYQRSTLMNOFGHJABCD";
static const double irish_lat0 = 53.5;
static const double irish_lon0 = -8.0;
static const double irish_k0 = 1.000035;

static const double irish_f_e = 200000.;
static const double irish_f_n = 250000.;

/* geog_to_irish:
 * Convert geographical coordinates in the given datum into an Irish Grid zone
 * and eastings/northings.
 */
int
geog_to_irish (double lat, double lon, double *x, double *y, int datum)
{
  geog_to_tm (lat, lon, irish_lat0, irish_lon0, irish_k0, x, y, datum);

  /* add false easting and northing and round to nearest meter */
  *x = *x + irish_f_e;
  *y = *y + irish_f_n;

  /* check for invalid range */
  if (*x < 0.0 || *x > 4.0e5 || *y < 0.0 || *y > 5e5)
    {
      *x = 0.0;
      *y = 0.0;

      return 0;
    }

  return 1;
}

/* irish_to_geog:
 * Convert Irish Grid coordinates into geographical coordinates in the given
 * datum.
 */
int
irish_to_geog (double x, double y, double *lat, double *lon, int datum)
{
  int m;

  x -= irish_f_e;
  y -= irish_f_n;

  tm_to_geog (x, y, irish_lat0, irish_lon0, irish_k0, lat, lon, datum);
//  geog_to_tm(x, y, irish_lat0, irish_lon0, irish_k0, lat, lon);

  return 1;
}
