#include "project.h"

#include "coords/libcoords.h"

#define CORRECT_X 0
#define CORRECT_Y -4658

/* Utility functions*/

static ImageCoord
iosxytoiic (IOSXY xy)
{
  return ImageCoord (xy.x, xy.y);
}

static IOSXY
iictoiosxy (ImageCoord ic)
{
  return IOSXY (ic.x, ic.y);
}

static WGS84
lltowgs84 (LatLon ll)
{
  return WGS84 (ll.lat, ll.lon);
}

static LatLon
wgs84toll (WGS84 w)
{
  return LatLon (w.lat, w.lon);
}

/* IOS Image coordinate system */

string IosImgCoordSystem::name (void)
{
  string ret;
  char buf[1024];
  sprintf (buf, "Irish Ordinance Survey grid");
  ret = buf;
  return ret;
}

ImageCoord IosImgCoordSystem::lltoic (LatLon ll)
{
  IOSXY xy = wgs84_to_iosxy (lltowgs84 (ll));
  return ImageCoord (iosxytoiic (xy));
}

LatLon IosImgCoordSystem::ictoll (ImageCoord ic)
{
  WGS84
  w =
    iosxy_to_wgs84 (iictoiosxy (ic));
  return LatLon (w.lat, w.lon);
}

/* Utility functions to handle those wacky letters*/

static int
letter_to_x (char c)
{
  switch (c | 32) {
    case 'a':
    case 'f':
    case 'l':
    case 'q':
    case 'v':
      return 0;
    case 'b':
    case 'g':
    case 'm':
    case 'r':
    case 'w':
      return 1;
    case 'c':
    case 'h':
    case 'n':
    case 's':
    case 'x':
      return 2;
    case 'd':
    case 'j':
    case 'o':
    case 't':
    case 'y':
      return 3;
  }
}


static int
letter_to_y (char c)
{
  switch (c | 32) {
    case 'a':
    case 'b':
    case 'c':
    case 'd':
      return 4;
    case 'f':
    case 'g':
    case 'h':
    case 'j':
      return 3;
    case 'l':
    case 'm':
    case 'n':
    case 'o':
      return 2;
    case 'q':
    case 'r':
    case 's':
    case 't':
      return 1;
    case 'v':
    case 'w':
    case 'x':
    case 'y':
      return 0;
  }
}

static char
xy_to_letter (int x, int y)
{
  char tbl[5][4] = { {'V', 'W', 'X', 'Y'},
                     {'Q', 'R', 'S', 'T'},
                     {'L', 'M', 'N', 'O'},
                     {'F', 'G', 'H', 'J'},
                     {'A', 'B', 'C', 'D'}
                   };

  if (x < 0)
    return ' ';
  if (y < 0)
    return ' ';
  if (x > 3)
    return ' ';
  if (y > 4)
    return ' ';

  return tbl[y][x];
}


class IOSGRID
{
  static IOSGRID IOSGRID::iosxy_to_iosgrid (IOSXY xy);

public:
  char zone;
  double e;
  double n;

  IOSGRID () {} ;

  IOSGRID (IOSXY xy)
  {
    *this = iosxy_to_iosgrid (xy);
  }
};


static IOSXY iosgrid_to_iosxy (IOSGRID g)
{
  IOSXY r (g.e, g.n);


  r.x += (double) (100000 * letter_to_x (g.zone));
  r.y += (double) (100000 * letter_to_y (g.zone));

  return r;
}


static IOSGRID iosxy_to_iosgrid (IOSXY xy)
{
  IOSGRID r;
  int x, y;

  x = (int) (xy.x / 100000.0);
  y = (int) (xy.y / 100000.0);

  r.zone = xy_to_letter (x, y);

  xy.x -= (double) (x * 100000);
  xy.y -= (double) (y * 100000);

  r.e = xy.x;
  r.n = xy.y;

  return r;
}



/* IOS coordinate system */

string IosCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "Irish Ordinance Survery grid references");

  return buf;
}

string IosCoordSystem::lltocs (LatLon ll)
{
  IOSGRID g = iosxy_to_iosgrid (wgs84_to_iosxy (lltowgs84 (ll)));
  char buf[1024];

  sprintf (buf, "%c%05d%05d", g.zone, (int) (g.e + 0.5), (int) (g.n + 0.5));

  return buf;

}


LatLon IosCoordSystem::cstoll (string s)
{
  LatLon ll;
  IOSGRID g;
  int i = s.length ();
  int
  j;

  if (!(i % 2))
    return ll;

  g.zone = s[0];

  i--;
  i /= 2;

  g.e = g.n = 0;

  for (j = 0; j < i; ++j) {
    g.e *= 10.0;
    g.e += (double) s[j + 1] - '0';
    g.n *= 10.0;
    g.n += (double) s[i + j + 1] - '0';
  }
  for (; j < 5; ++j) {
    g.n *= 10.0;
    g.e *= 10.0;
  }

  WGS84 w = iosxy_to_wgs84 (iosgrid_to_iosxy (g));
  return LatLon (w.lat, w.lon);
}


/* Map object */

class IosTileDescription: public TileDescription
{
public:
  IosTileDescription (IOSXY ixy, int iw, int ih)
  {
    xy = ixy;
    w = iw;
    h = ih;
  }
  int issame (TileDescription * it1, TileDescription * it2)
  {
    IosTileDescription * t1 = (IosTileDescription *) it1;
    IosTileDescription * t2 = (IosTileDescription *) it2;

    if (t1->xy.x != t2->xy.x)
      return 0;
    if (t1->xy.y != t2->xy.y)
      return 0;
    if (t1->w != t2->w)
      return 0;
    if (t1->h != t2->h)
      return 0;
    return 1;
  };
  IOSXY xy;
  int w;
  int h;
};

class IosMap: public Map
{
public:
  string name (void);
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort);
  ~IosMap () {} ;

  double offx;
  double offy;
  double scalex;
  double scaley;
  IosImgCoordSystem iics;

  TileCache tc;

  virtual Tile * load_tile (IosTileDescription * td, int effort) = 0;
};

#define O10KX 6491
#define O10KY  1955
#define S10KX 7800
#define S10KY 5780

class IosMap10k: public IosMap
{
public:
  IosMap10k ()
  {
    offx = O10KX;
    offy = O10KY;
    scalex = S10KX;
    scaley = S10KY;
    preferedics = new IosImgCoordSystem;
    preferedscale = 20.0;
    tiled = 1;
  };
  Tile * load_tile (IosTileDescription * td, int effort)
  {
    if ((effort == RENDER_LAZY) && ((td->w * td->h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      return NULL;
    }
    if ((td->w <= 0) || (td->h <= 0))
      return NULL;

    int x = (int) (td->xy.x + .5);
    int y = (int) (td->xy.y + .5);

    x -= O10KX;
    x = x / S10KX;
    x *= S10KX;
    x += O10KX;

    y -= O10KY;
    y = y / S10KY;
    y *= S10KY;
    y += O10KY;

    char fn[1024];
    sprintf (fn, "%s/%06d/%06d.%06d.jpg", getenv ("IOSPATH"), x, x, y);

    // printf("Loading from %s\n", fn);

    GdkPixbuf * b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;
    GdkPixbuf * sb = gdk_pixbuf_scale_simple (b, td->w, td->h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    if (!sb)
      return NULL;

    Tile * t = new Tile;
    t->td = new IosTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
  ~IosMap10k ()
  {
    delete preferedics;
  };
};



Map * create_iosmap ()
{
  return new IosMap10k;
}

string IosMap::name (void)
{
  char buf[1024];
  sprintf (buf, "Irish Ordinance Survey Map scale %.1fkm",
           (double) (scalex / 10000));
  return buf;
}

#define BOUND(v,l,h) v= (((v)>(h)) ? (h):( ((v)<(l))?(l):v ))


void IosMap::render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p , int effort)
{

  IOSXY ll = iictoiosxy (iics.lltoic (vcs->vctoll (ViewCoord (p->x, p->y + p->h))));
  IOSXY ur = iictoiosxy (iics.lltoic (vcs->vctoll (ViewCoord (p->x + p->w, p->y))));

  ll.x -= CORRECT_X;
  ur.x -= CORRECT_X;
  ll.y -= CORRECT_Y;
  ur.y -= CORRECT_Y;


  ll.x = round_down (ll.x - offx, scalex) + offx;
  ll.y = round_down (ll.y - offy, scaley) + offy;



  IOSXY xy;
  xy.y = (ur.y - ll.y) / scaley;
  xy.x = (ur.x - ll.x) / scalex;
  xy.y *= xy.x;

  if ((effort == RENDER_LAZY) && (xy.y > MAXSUBTILES)) {
    printf ("MAXSUBTILES stitching limit exceeded tile not done\n");
    return ;
  }


  for (xy.y = ll.y; xy.y < ur.y; xy.y += scaley) {
    for (xy.x = ll.x; xy.x < ur.x; xy.x += scalex) {


      ViewCoord vc_ul = vcs->lltovc (iics.ictoll (iosxytoiic (IOSXY (xy.x + CORRECT_X, xy.y + scaley + CORRECT_Y))));
      ViewCoord vc_lr = vcs->lltovc (iics. ictoll (iosxytoiic (IOSXY (xy.x + scalex + CORRECT_X, xy.y + CORRECT_Y))));
      int tw = (int) ((vc_lr.x - vc_ul.x) + .5);
      int th = (int) ((vc_lr.y - vc_ul.y) + .5);


      IosTileDescription td (xy, tw, th);

      Tile *t = tc.find (&td);
      if (!t)
        t = load_tile (&td, effort);

      if ((t) && (!t->b))
        FATALERROR ("Fish");

      double dsx = (int) (vc_ul.x + .5);
      double dsy = (int) (vc_ul.y + .5);

      BOUND (vc_ul.x, p->x, p->x + p->w);
      BOUND (vc_ul.y, p->y, p->y + p->h);
      BOUND (vc_lr.x, p->x, p->x + p->w);
      BOUND (vc_lr.y, p->y, p->y + p->h);

      int x = (int) (vc_ul.x + .5);
      int y = (int) (vc_ul.y + .5);

      int sx = (int) (0.5 + vc_ul.x - dsx);
      int sy = (int) (0.5 + vc_ul.y - dsy);

      int w = (int) ((vc_lr.x - vc_ul.x) + .5);
      int h = (int) ((vc_lr.y - vc_ul.y) + .5);

      if (w && h) {
        if (t) {
          if ((w + sx) > tw)
            w -= (w + sx) - tw;
          if ((h + sy) > th)
            h -= (h + sy) - th;

          d->render (t->b, sx, sy, x, y, w, h, x, y);
        } else {
          if ((w > 0) || (h > 0))
            d->zap (x, y, w, h);
        }
      }

    }
  }
}
