#include <sys/stat.h>

#include "project.h"

#define FONTSIZE 12
#define LINEWIDTH 3

extern int markup_p_line;
extern int markup_p_parse(void);
char *markup_p_name;
static FILE *markup_p_file;
static class MarkupList *markup_p_list;

static int oddonly(int i)
{
  if (i&1)
    return i;
  return 0;
}




class Markup;

class MarkupList
{
public:
  Markup *head, *tail;
  void add(Markup *m);
  char *path;
  time_t mtime;
  MarkupList(char *);
  void Update(void);
  ~MarkupList();
};

class Markup
{
public:
  Markup *prev, *next;
  MarkupList *parent;
  virtual void draw(GisDrawable *d, ViewCoordSystem *vcs) = 0;

  ~Markup()
  {
    if (prev) {
      prev->next = next;
    } else {
      parent->head = next;
    }
    if (next) {
      next->prev = prev;
    } else {
      parent->tail = prev;
    }
  };
};


class MIcon: public Markup
{
public:
  LatLon l;
  char *path;
  GdkPixbuf *b;
  int w, h;

  MIcon(char *p, LatLon ul)
  {
    path = strdup(p);
    b = gdk_pixbuf_new_from_file (p);
    if (!b)
      fprintf(stderr, "MIcon couldn't load %s\n", p);
    w = gdk_pixbuf_get_width(b);
    h = gdk_pixbuf_get_height(b);
    l = ul;
  };

  void draw(GisDrawable *d, ViewCoordSystem *vcs)
  {
    ViewCoord vc = vcs->lltovc(l);
    int x = ((int) vc.x) - (w >> 1);
    int y = ((int) vc.y) - (h >> 1);

    d->render(b, 0, 0, x, y, w, h, x, y);
  };

  ~MIcon()
  {
    gdk_pixbuf_finalize (b);
    free(path);
  };
};



class MText: public Markup
{
public:
  LatLon l;
  char *text;
  GisColor c;
  int s;

  MText(char *p, LatLon ul, GisColor uc, int is)
  {
    text = strdup(p);
    l = ul;
    c = uc;
    if (is)
      s = is;
    else
      s = FONTSIZE;
  };

  void draw(GisDrawable *d, ViewCoordSystem *vcs)
  {
    ViewCoord vc = vcs->lltovc(l);
    int x = (int) vc.x;
    int y = (int) vc.y;

    d->text(text, x, y, c, s);
  };

  ~MText()
  {
    free(text);
  };
};


class MELine: public Markup
{
public:
  LatLon l1, l2;
  GisColor c;
  int w;

  MELine(LatLon ul1, LatLon ul2, GisColor uc, int iw)
  {
    l1 = ul1;
    l2 = ul2;
    c = uc;
    w = iw;
  };

  void draw(GisDrawable *d, ViewCoordSystem *vcs)
  {
    ViewCoord vc1 = vcs->lltovc(l1);
    ViewCoord vc2 = vcs->lltovc(l2);
    ViewCoord vc3;

    //FIXME do this properly with GC intersections of the View.

    double dx = vc2.x - vc1.x;
    double dy = vc2.y - vc1.y;

    double x1, y1, x2, y2;

    if (fabs(dy) > fabs(dx)) {
      if (dy < 0) {
        vc3 = vc2;
        vc2 = vc1;
        vc1 = vc3;
        dx = -dx;
        dy = -dy;
      }

      y1 = 0;
      x1 = vc1.x - (((vc1.y - y1) / dy) * dx);

      y2 = d->h;
      x2 = vc2.x - (((vc2.y - y2) / dy) * dx);

    } else {
      if (dx < 0) {
        vc3 = vc2;
        vc2 = vc1;
        vc1 = vc3;
        dx = -dx;
        dy = -dy;
      }

      x1 = 0;
      y1 = vc1.y - (((vc1.x - x1) / dx) * dy);

      x2 = d->w;
      y2 = vc2.y - (((vc2.x - x2) / dx) * dy);

    }

    d->line((int) x1, (int) y1, (int) x2, (int) y2, c, w);

  };

};


class MLine: public Markup
{
public:
  LatLon l1, l2;
  GisColor c;
  int w;

  MLine(LatLon ul1, LatLon ul2, GisColor uc, int iw)
  {
    l1 = ul1;
    l2 = ul2;
    c = uc;
    w = iw;
  };

  void draw(GisDrawable *d, ViewCoordSystem *vcs)
  {
    ViewCoord vc1 = vcs->lltovc(l1);
    ViewCoord vc2 = vcs->lltovc(l2);

    int x1 = (int) vc1.x;
    int x2 = (int) vc2.x;
    int y1 = (int) vc1.y;
    int y2 = (int) vc2.y;

    d->line(x1, y1, x2, y2, c, w);
  };

};

MarkupList::MarkupList(char *p)
{
  path = strdup(p);
  head = 0;
  tail = 0;
  mtime = (time_t) 0;

  Update();
}

int markup_p_read(char *buf, int num)
{
  return fread(buf, 1, num, markup_p_file);
}

void markup_p_icon(double lat, double lon, char *fname)
{
  markup_p_list->add(new MIcon(fname, LatLon(lat, lon)));
}

void markup_p_ntext(double lat, double lon, char *text, char *color, double size)
{
  markup_p_list->add(new MText(text, LatLon(lat, lon), GisColor(color), (int) (size + .5)));
}

void markup_p_nline(double l1, double l2, double l3, double l4, char *color, double w)
{
  markup_p_list->add( new MLine(LatLon(l1, l2), LatLon(l3, l4), GisColor(color), (int) w));
}

void markup_p_eline(double l1, double l2, double l3, double l4, char *color, double w)
{
  markup_p_list->add( new MELine(LatLon(l1, l2), LatLon(l3, l4), GisColor(color), (int) w));
}


void MarkupList::Update(void)
{
  struct stat stbuf;
  char buf[1024];

  if (stat(path, &stbuf))
    return ;
  if (mtime == stbuf.st_mtime)
    return ;

  mtime = stbuf.st_mtime;

  while (head)
    delete head;

  markup_p_name = path;
  markup_p_file = fopen(path, "r");
  markup_p_list = this;

  if (!markup_p_file)
    return ;

  markup_p_line = 0;

  markup_p_parse();

  fclose(markup_p_file);


}


MarkupList::~MarkupList()
{
  while (head)
    delete head;
  free(path);
};


void MarkupList::add(Markup *m)
{
  if (tail) {
    tail->next = m;
    m->prev = tail;
    m->prev->next = m;
    m->next = 0;
    tail = m;
  } else {
    head = m;
    tail = m;
    m->next = 0;
    m->prev = 0;
  }
  m->parent = this;
};


static MarkupList *mul = 0;

static void check_markup(void)
{
  if (!mul)
    mul = new MarkupList("markup.dat");
  else
    mul->Update();


}



class MarkupMap: public Map
{
public:
  string name(void)
  {
    return string("Markup");
  }
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
  {

    check_markup();

    if (!mul)
      return ;

    Markup *m = mul->head;
    while (m) {
      m->draw(d, vcs);

      m = m->next;
    }

  }
  MarkupMap()
  {
    tiled = 0;
    preferedics = 0;
  };
};

Map * create_markupmap (void)
{
  return new MarkupMap;
}

